package dev.pfaff.altargui.util;

import dev.pfaff.altargui.function.FallibleRunnable;
import org.jspecify.annotations.Nullable;

import java.util.concurrent.Callable;
import java.util.concurrent.CompletionException;

import static dev.pfaff.altargui.internal.Config.TRACE_THREAD_FUTURE_SPAWN_POINT;
import static dev.pfaff.altargui.util.ThrowUtil.throwUnchecked;

public final class ThreadFuture<V> implements BasicFuture<V> {
	private final Thread thread;
	@Nullable
	private final SpawnTrace spawnTrace;

	private V value;
	@Nullable
	private Throwable throwable;

	private ThreadFuture(FallibleRunnable runnable) {
		this.thread = Thread.startVirtualThread(() -> {
			try {
				runnable.run();
			} catch (Throwable e) {
				this.throwable = e;
			}
		});
		this.spawnTrace = TRACE_THREAD_FUTURE_SPAWN_POINT ? new SpawnTrace() : null;
	}

	private ThreadFuture(Callable<V> callable) {
		this.thread = Thread.startVirtualThread(() -> {
			try {
				this.value = callable.call();
			} catch (Throwable e) {
				this.throwable = e;
			}
		});
		this.spawnTrace = TRACE_THREAD_FUTURE_SPAWN_POINT ? new SpawnTrace() : null;
	}

	public static ThreadFuture<Void> spawnVirtualTask(FallibleRunnable runnable) {
		return new ThreadFuture<>(runnable);
	}

	public static <V> ThreadFuture<V> spawnVirtualTask(Callable<V> callable) {
		return new ThreadFuture<>(callable);
	}

	public V join() throws CompletionException, CancellationException {
		try {
			this.thread.join();
		} catch (InterruptedException e) {
			throw throwUnchecked(e);
		}
		if (this.throwable != null) {
			RuntimeException e;
			if (this.throwable instanceof CancellationException) {
				e = new CancellationException("Thread future cancelled", this.throwable);
			} else {
				e = new CompletionException("Thread future completed with exception", this.throwable);
			}
			if (TRACE_THREAD_FUTURE_SPAWN_POINT) {
				assert this.spawnTrace != null;
				e.addSuppressed(this.spawnTrace);
			}
			throw e;
		}
		return this.value;
	}

	@Override
	public void cancel() {
		this.thread.interrupt();
	}

	public boolean isAlive() {
		return this.thread.isAlive();
	}

	public Thread thread() {
		return this.thread;
	}

	private static final class SpawnTrace extends Throwable {
	}
}
